""".. Ignore pydocstyle D400.

==========
Data Purge
==========

"""
import logging
import os
import shutil

from django.db.models import Q

from resolwe.flow.models import Data, DataLocation, Storage
from resolwe.flow.utils import iterate_fields
from resolwe.utils import BraceMessage as __

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def get_purge_files(root, output, output_schema, descriptor, descriptor_schema):
    """Get files to purge."""
    def remove_file(fn, paths):
        """From paths remove fn and dirs before fn in dir tree."""
        while fn:
            for i in range(len(paths) - 1, -1, -1):
                if fn == paths[i]:
                    paths.pop(i)
            fn, _ = os.path.split(fn)

    def remove_tree(fn, paths):
        """From paths remove fn and dirs before or after fn in dir tree."""
        for i in range(len(paths) - 1, -1, -1):
            head = paths[i]
            while head:
                if fn == head:
                    paths.pop(i)
                    break
                head, _ = os.path.split(head)

        remove_file(fn, paths)

    def subfiles(root):
        """Extend unreferenced list with all subdirs and files in top dir."""
        subs = []
        for path, dirs, files in os.walk(root, topdown=False):
            path = path[len(root) + 1:]
            subs.extend(os.path.join(path, f) for f in files)
            subs.extend(os.path.join(path, d) for d in dirs)
        return subs

    unreferenced_files = subfiles(root)

    remove_file('jsonout.txt', unreferenced_files)
    remove_file('stderr.txt', unreferenced_files)
    remove_file('stdout.txt', unreferenced_files)

    meta_fields = [
        [output, output_schema],
        [descriptor, descriptor_schema]
    ]

    for meta_field, meta_field_schema in meta_fields:
        for field_schema, fields in iterate_fields(meta_field, meta_field_schema):
            if 'type' in field_schema:
                field_type = field_schema['type']
                field_name = field_schema['name']

                # Remove basic:file: entries
                if field_type.startswith('basic:file:'):
                    remove_file(fields[field_name]['file'], unreferenced_files)

                # Remove list:basic:file: entries
                elif field_type.startswith('list:basic:file:'):
                    for field in fields[field_name]:
                        remove_file(field['file'], unreferenced_files)

                # Remove basic:dir: entries
                elif field_type.startswith('basic:dir:'):
                    remove_tree(fields[field_name]['dir'], unreferenced_files)

                # Remove list:basic:dir: entries
                elif field_type.startswith('list:basic:dir:'):
                    for field in fields[field_name]:
                        remove_tree(field['dir'], unreferenced_files)

                # Remove refs entries
                if field_type.startswith('basic:file:') or field_type.startswith('basic:dir:'):
                    for ref in fields[field_name].get('refs', []):
                        remove_tree(ref, unreferenced_files)

                elif field_type.startswith('list:basic:file:') or field_type.startswith('list:basic:dir:'):
                    for field in fields[field_name]:
                        for ref in field.get('refs', []):
                            remove_tree(ref, unreferenced_files)

    return set([os.path.join(root, filename) for filename in unreferenced_files])


def location_purge(location_id, delete=False, verbosity=0):
    """Print and conditionally delete files not referenced by meta data.

    :param location_id: Id of the
        :class:`~resolwe.flow.models.DataLocation` model that data
        objects reference to.
    :param delete: If ``True``, then delete unreferenced files.
    """
    try:
        location = DataLocation.objects.get(id=location_id)
    except DataLocation.DoesNotExist:
        logger.warning("Data location does not exist", extra={'location_id': location_id})
        return

    unreferenced_files = set()
    purged_data = Data.objects.none()
    referenced_by_data = location.data.exists()
    if referenced_by_data:
        if location.data.exclude(status__in=[Data.STATUS_DONE, Data.STATUS_ERROR]).exists():
            return

        # Perform cleanup.
        purge_files_sets = list()
        purged_data = location.data.all()
        for data in purged_data:
            purge_files_sets.append(get_purge_files(
                location.get_path(),
                data.output,
                data.process.output_schema,
                data.descriptor,
                getattr(data.descriptor_schema, 'schema', [])
            ))

        intersected_files = set.intersection(*purge_files_sets) if purge_files_sets else set()
        unreferenced_files.update(intersected_files)
    else:
        # Remove data directory.
        unreferenced_files.add(location.get_path())
        unreferenced_files.add(location.get_runtime_path())

    if verbosity >= 1:
        # Print unreferenced files
        if unreferenced_files:
            logger.info(__("Unreferenced files for location id {} ({}):", location_id, len(unreferenced_files)))
            for name in unreferenced_files:
                logger.info(__("  {}", name))
        else:
            logger.info(__("No unreferenced files for location id {}", location_id))

    # Go through unreferenced files and delete them.
    if delete:
        for name in unreferenced_files:
            if os.path.isfile(name) or os.path.islink(name):
                os.remove(name)
            elif os.path.isdir(name):
                shutil.rmtree(name)

        location.purged = True
        location.save()

        if not referenced_by_data:
            location.delete()


def _location_purge_all(delete=False, verbosity=0):
    """Purge all data locations."""
    if DataLocation.objects.exists():
        for location in DataLocation.objects.filter(Q(purged=False) | Q(data=None)):
            location_purge(location.id, delete, verbosity)
    else:
        logger.info("No data locations")


def _storage_purge_all(delete=False, verbosity=0):
    """Purge unreferenced storages."""
    orphaned_storages = Storage.objects.filter(data=None)

    if verbosity >= 1:
        if orphaned_storages.exists():
            logger.info(__("Unreferenced storages ({}):", orphaned_storages.count()))
            for storage_id in orphaned_storages.values_list('id', flat=True):
                logger.info(__("  {}", storage_id))
        else:
            logger.info("No unreferenced storages")

    if delete:
        orphaned_storages.delete()


def purge_all(delete=False, verbosity=0):
    """Purge all data locations."""
    _location_purge_all(delete, verbosity)
    _storage_purge_all(delete, verbosity)



"""Channels consumers for Resolwe Flow."""
import logging

from channels.consumer import SyncConsumer

from resolwe.flow.utils.purge import location_purge
from resolwe.utils import BraceMessage as __

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class PurgeConsumer(SyncConsumer):
    """Purge consumer."""

    def purge_run(self, event):
        """Run purge for the object with ``location_id`` specified in ``event`` argument."""
        location_id = event['location_id']
        verbosity = event['verbosity']

        try:
            logger.info(__("Running purge for location id {}.", location_id))
            location_purge(location_id=location_id, delete=True, verbosity=verbosity)
        except Exception:  # pylint: disable=broad-except
            logger.exception("Error while purging location.", extra={'location_id': location_id})



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-10-01 03:55
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0023_process_entity_2'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='process',
            name='flow_collection',
        ),
    ]



"""Resolwe entity serializer."""
from rest_framework import serializers

from resolwe.flow.models import Entity

from .collection import CollectionSerializer


class EntitySerializer(CollectionSerializer):
    """Serializer for Entity."""

    collections = serializers.PrimaryKeyRelatedField(many=True, read_only=True)

    class Meta(CollectionSerializer.Meta):
        """EntitySerializer Meta options."""

        model = Entity
        fields = CollectionSerializer.Meta.fields + (
            'collections',
            'descriptor_completed',
            'duplicated',
            'type',
        )

    def get_data(self, entity):
        """Return serialized list of data objects on entity that user has `view` permission on."""
        data = self._filter_queryset('view_data', entity.data.all())

        return self._serialize_data(data)



"""An execution engine for Python processes."""
import copy
import inspect
import json
import os
import shlex
import shutil

from resolwe.flow.execution_engines.base import BaseExecutionEngine
from resolwe.flow.execution_engines.exceptions import ExecutionError
from resolwe.flow.models.utils import hydrate_input_references, hydrate_input_uploads
from resolwe.process.parser import SafeParser

PYTHON_RUNTIME_DIRNAME = 'python_runtime'
PYTHON_RUNTIME_ROOT = '/'
PYTHON_RUNTIME_VOLUME = os.path.join(PYTHON_RUNTIME_ROOT, PYTHON_RUNTIME_DIRNAME)
PYTHON_PROGRAM_ROOT = '/'
PYTHON_PROGRAM_FILENAME = 'python_process.py'
PYTHON_PROGRAM_VOLUME = os.path.join(PYTHON_PROGRAM_ROOT, PYTHON_PROGRAM_FILENAME)
PYTHON_INPUTS_FILENAME = 'inputs.json'
PYTHON_INPUTS_ROOT = '/'
PYTHON_INPUTS_VOLUME = os.path.join(PYTHON_INPUTS_ROOT, PYTHON_INPUTS_FILENAME)


class ExecutionEngine(BaseExecutionEngine):
    """An execution engine that outputs bash programs."""

    name = 'python'

    def discover_process(self, path):
        """Perform process discovery in given path.

        This method will be called during process registration and
        should return a list of dictionaries with discovered process
        schemas.
        """
        if not path.lower().endswith('.py'):
            return []

        parser = SafeParser(open(path).read())
        processes = parser.parse()
        return [process.to_schema() for process in processes]

    def evaluate(self, data):
        """Evaluate the code needed to compute a given Data object."""
        return 'PYTHONPATH="{runtime}" python3 -m resolwe.process {program} --slug {slug} --inputs {inputs}'.format(
            runtime=PYTHON_RUNTIME_VOLUME,
            program=PYTHON_PROGRAM_VOLUME,
            slug=shlex.quote(data.process.slug),
            inputs=PYTHON_INPUTS_VOLUME,
        )

    def prepare_runtime(self, runtime_dir, data):
        """Prepare runtime directory."""
        # Copy over Python process runtime (resolwe.process).
        import resolwe.process as runtime_package

        src_dir = os.path.dirname(inspect.getsourcefile(runtime_package))
        dest_package_dir = os.path.join(runtime_dir, PYTHON_RUNTIME_DIRNAME, 'resolwe', 'process')
        shutil.copytree(src_dir, dest_package_dir)
        os.chmod(dest_package_dir, 0o755)

        # Write python source file.
        source = data.process.run.get('program', '')
        program_path = os.path.join(runtime_dir, PYTHON_PROGRAM_FILENAME)
        with open(program_path, 'w') as file:
            file.write(source)
        os.chmod(program_path, 0o755)

        # Write serialized inputs.
        inputs = copy.deepcopy(data.input)
        hydrate_input_references(inputs, data.process.input_schema)
        hydrate_input_uploads(inputs, data.process.input_schema)
        inputs_path = os.path.join(runtime_dir, PYTHON_INPUTS_FILENAME)

        # XXX: Skip serialization of LazyStorageJSON. We should support
        # LazyStorageJSON in Python processes on the new communication protocol
        def default(obj):
            """Get default value."""
            class_name = obj.__class__.__name__
            if class_name == 'LazyStorageJSON':
                return ''

            raise TypeError(f'Object of type {class_name} is not JSON serializable')

        with open(inputs_path, 'w') as file:
            json.dump(inputs, file, default=default)

        # Generate volume maps required to expose needed files.
        volume_maps = {
            PYTHON_RUNTIME_DIRNAME: PYTHON_RUNTIME_VOLUME,
            PYTHON_PROGRAM_FILENAME: PYTHON_PROGRAM_VOLUME,
            PYTHON_INPUTS_FILENAME: PYTHON_INPUTS_VOLUME,
        }

        return volume_maps



"""Iterator utils."""
import collections


def iterate_fields(fields, schema, path_prefix=None):
    """Iterate over all field values sub-fields.

    This will iterate over all field values. Some fields defined in the schema
    might not be visited.

    :param fields: field values to iterate over
    :type fields: dict
    :param schema: schema to iterate over
    :type schema: dict
    :return: (field schema, field value)
    :rtype: tuple

    """
    if path_prefix is not None and path_prefix != '' and path_prefix[-1] != '.':
        path_prefix += '.'

    schema_dict = {val['name']: val for val in schema}
    for field_id, properties in fields.items():
        path = '{}{}'.format(path_prefix, field_id) if path_prefix is not None else None
        if field_id not in schema_dict:
            raise KeyError("Field definition ({}) missing in schema".format(field_id))
        if 'group' in schema_dict[field_id]:
            for rvals in iterate_fields(properties, schema_dict[field_id]['group'], path):
                yield rvals if path_prefix is not None else rvals[:2]
        else:
            rvals = (schema_dict[field_id], fields, path)
            yield rvals if path_prefix is not None else rvals[:2]


def iterate_schema(fields, schema, path_prefix=''):
    """Iterate over all schema sub-fields.

    This will iterate over all field definitions in the schema. Some field v
    alues might be None.

    :param fields: field values to iterate over
    :type fields: dict
    :param schema: schema to iterate over
    :type schema: dict
    :param path_prefix: dot separated path prefix
    :type path_prefix: str
    :return: (field schema, field value, field path)
    :rtype: tuple

    """
    if path_prefix and path_prefix[-1] != '.':
        path_prefix += '.'

    for field_schema in schema:
        name = field_schema['name']
        if 'group' in field_schema:
            for rvals in iterate_schema(fields[name] if name in fields else {},
                                        field_schema['group'], '{}{}'.format(path_prefix, name)):
                yield rvals
        else:
            yield (field_schema, fields, '{}{}'.format(path_prefix, name))


def iterate_dict(container, exclude=None, path=None):
    """Iterate over a nested dictionary.

    The dictionary is iterated over in a depth first manner.

    :param container: Dictionary to iterate over
    :param exclude: Optional callable, which is given key and value as
        arguments and may return True to stop iteration of that branch
    :return: (path, key, value) tuple
    """
    if path is None:
        path = []

    for key, value in container.items():
        if callable(exclude) and exclude(key, value):
            continue

        if isinstance(value, collections.Mapping):
            for inner_path, inner_key, inner_value in iterate_dict(value, exclude=exclude, path=path + [key]):
                yield inner_path, inner_key, inner_value

        yield path, key, value



"""Elastic search indexes for Resolwe models."""

from .collection import CollectionDocument, CollectionIndex
from .data import DataDocument, DataIndex
from .entity import EntityDocument, EntityIndex



""".. Ignore pydocstyle D400.

========
Listener
========

Standalone Redis client used as a contact point for executors.

.. autoclass:: resolwe.flow.managers.listener.ExecutorListener
    :members:

"""

import asyncio
import json
import logging
import math
import os
import time
import traceback

import aioredis
from asgiref.sync import async_to_sync
from channels.db import database_sync_to_async
from channels.exceptions import ChannelFull
from channels.layers import get_channel_layer

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.models import Count
from django.urls import reverse

from django_priority_batch import PrioritizedBatcher

from resolwe.flow.models import Data, DataDependency, Entity, Process
from resolwe.flow.protocol import CHANNEL_PURGE_WORKER, TYPE_PURGE_RUN
from resolwe.flow.utils import dict_dot, iterate_fields, stats
from resolwe.permissions.utils import copy_permissions
from resolwe.test.utils import is_testing
from resolwe.utils import BraceMessage as __

from . import consumer, state
from .protocol import ExecutorProtocol, WorkerProtocol

if settings.USE_TZ:
    from django.utils.timezone import now  # pylint: disable=ungrouped-imports
else:
    import datetime
    now = datetime.datetime.now  # pylint: disable=invalid-name

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ExecutorListener:
    """The contact point implementation for executors."""

    def __init__(self, *args, **kwargs):
        """Initialize attributes.

        :param host: Optional. The hostname where redis is running.
        :param port: Optional. The port where redis is running.
        """
        super().__init__()

        # The Redis connection object.
        self._redis = None
        self._redis_params = kwargs.get('redis_params', {})

        # Running coordination.
        self._should_stop = False
        self._runner_coro = None

        # The verbosity level to pass around to Resolwe utilities
        # such as location_purge.
        self._verbosity = kwargs.get('verbosity', 1)

        # Statistics about how much time each event needed for handling.
        self.service_time = stats.NumberSeriesShape()

        # Statistics about the number of events handled per time interval.
        self.load_avg = stats.SimpleLoadAvg([60, 5 * 60, 15 * 60])

        # Timestamp of last critical load error and level, for throttling.
        self.last_load_log = -math.inf
        self.last_load_level = 0

    async def _make_connection(self):
        """Construct a connection to Redis."""
        return await aioredis.create_redis(
            'redis://{}:{}'.format(
                self._redis_params.get('host', 'localhost'),
                self._redis_params.get('port', 6379)
            ),
            db=int(self._redis_params.get('db', 1))
        )

    async def _call_redis(self, meth, *args, **kwargs):
        """Perform a Redis call and handle connection dropping."""
        while True:
            try:
                if not self._redis:
                    self._redis = await self._make_connection()
                return await meth(self._redis, *args, **kwargs)
            except aioredis.RedisError:
                logger.exception("Redis connection error")
                if self._redis:
                    self._redis.close()
                    await self._redis.wait_closed()
                    self._redis = None
                await asyncio.sleep(3)

    async def clear_queue(self):
        """Reset the executor queue channel to an empty state."""
        conn = await self._make_connection()
        try:
            script = """
                local keys = redis.call('KEYS', ARGV[1])
                redis.call('DEL', unpack(keys))
            """
            await conn.eval(
                script,
                keys=[],
                args=['*{}*'.format(settings.FLOW_MANAGER['REDIS_PREFIX'])],
            )
        finally:
            conn.close()

    async def __aenter__(self):
        """On entering a context, start the listener thread."""
        self._should_stop = False
        self._redis = await self._make_connection()
        self._runner_coro = asyncio.ensure_future(self.run())
        return self

    async def __aexit__(self, typ, value, trace):
        """On exiting a context, kill the listener and wait for it.

        .. note::

            Exceptions are all propagated.
        """
        await self._runner_coro
        self._redis.close()
        await self._redis.wait_closed()
        # Make sure the connection is cleaned up.
        self._redis = None

    def terminate(self):
        """Stop the standalone manager."""
        logger.info(__(
            "Terminating Resolwe listener on channel '{}'.",
            state.MANAGER_EXECUTOR_CHANNELS.queue
        ))
        self._should_stop = True

    def _queue_response_channel(self, obj):
        """Generate the feedback channel name from the object's id.

        :param obj: The Channels message object.
        """
        return '{}.{}'.format(state.MANAGER_EXECUTOR_CHANNELS.queue_response, obj[ExecutorProtocol.DATA_ID])

    # The handle_* methods are all Django synchronized, meaning they're run
    # in separate threads. Having this method be sync and calling async_to_sync
    # on rpush itself would mean reading self._redis from the sync thread,
    # which isn't very tidy. If it's async, it'll be called from the main
    # thread by the async_to_sync calls in handle_*.
    async def _send_reply(self, obj, reply):
        """Send a reply with added standard fields back to executor.

        :param obj: The original Channels message object to which we're
            replying.
        :param reply: The message contents dictionary. The data id is
            added automatically (``reply`` is modified in place).
        """
        reply.update({
            ExecutorProtocol.DATA_ID: obj[ExecutorProtocol.DATA_ID],
        })
        await self._call_redis(aioredis.Redis.rpush, self._queue_response_channel(obj), json.dumps(reply))

    def hydrate_spawned_files(self, exported_files_mapper, filename, data_id):
        """Pop the given file's map from the exported files mapping.

        :param exported_files_mapper: The dict of file mappings this
            process produced.
        :param filename: The filename to format and remove from the
            mapping.
        :param data_id: The id of the :meth:`~resolwe.flow.models.Data`
            object owning the mapping.
        :return: The formatted mapping between the filename and
            temporary file path.
        :rtype: dict
        """
        # JSON only has string dictionary keys, so the Data object id
        # needs to be stringified first.
        data_id = str(data_id)

        if filename not in exported_files_mapper[data_id]:
            raise KeyError("Use 're-export' to prepare the file for spawned process: {}".format(filename))

        export_fn = exported_files_mapper[data_id].pop(filename)

        if exported_files_mapper[data_id] == {}:
            exported_files_mapper.pop(data_id)

        return {'file_temp': export_fn, 'file': filename}

    def handle_update(self, obj, internal_call=False):
        """Handle an incoming ``Data`` object update request.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'update',
                    'data_id': [id of the :class:`~resolwe.flow.models.Data`
                               object this command changes],
                    'changeset': {
                        [keys to be changed]
                    }
                }

        :param internal_call: If ``True``, this is an internal delegate
            call, so a reply to the executor won't be sent.
        """
        data_id = obj[ExecutorProtocol.DATA_ID]
        changeset = obj[ExecutorProtocol.UPDATE_CHANGESET]
        if not internal_call:
            logger.debug(
                __("Handling update for Data with id {} (handle_update).", data_id),
                extra={
                    'data_id': data_id,
                    'packet': obj
                }
            )
        try:
            d = Data.objects.get(pk=data_id)
        except Data.DoesNotExist:
            logger.warning(
                "Data object does not exist (handle_update).",
                extra={
                    'data_id': data_id,
                }
            )

            if not internal_call:
                async_to_sync(self._send_reply)(obj, {ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_ERROR})

            async_to_sync(consumer.send_event)({
                WorkerProtocol.COMMAND: WorkerProtocol.ABORT,
                WorkerProtocol.DATA_ID: obj[ExecutorProtocol.DATA_ID],
                WorkerProtocol.FINISH_COMMUNICATE_EXTRA: {
                    'executor': getattr(settings, 'FLOW_EXECUTOR', {}).get('NAME', 'resolwe.flow.executors.local'),
                },
            })

            return

        if changeset.get('status', None) == Data.STATUS_ERROR:
            logger.error(
                __("Error occured while running process '{}' (handle_update).", d.process.slug),
                extra={
                    'data_id': data_id,
                    'api_url': '{}{}'.format(
                        getattr(settings, 'RESOLWE_HOST_URL', ''),
                        reverse('resolwe-api:data-detail', kwargs={'pk': data_id})
                    ),
                }
            )

        if d.status == Data.STATUS_ERROR:
            changeset['status'] = Data.STATUS_ERROR

        if not d.started:
            changeset['started'] = now()
        changeset['modified'] = now()

        for key, val in changeset.items():
            if key in ['process_error', 'process_warning', 'process_info']:
                # Trim process_* fields to not exceed max length of the database field.
                for i, entry in enumerate(val):
                    max_length = Data._meta.get_field(key).base_field.max_length  # pylint: disable=protected-access
                    if len(entry) > max_length:
                        val[i] = entry[:max_length - 3] + '...'

                getattr(d, key).extend(val)

            elif key != 'output':
                setattr(d, key, val)

        if 'output' in changeset:
            if not isinstance(d.output, dict):
                d.output = {}
            for key, val in changeset['output'].items():
                dict_dot(d.output, key, val)

        try:
            d.save(update_fields=list(changeset.keys()))
        except ValidationError as exc:
            logger.error(
                __(
                    "Validation error when saving Data object of process '{}' (handle_update):\n\n{}",
                    d.process.slug,
                    traceback.format_exc()
                ),
                extra={
                    'data_id': data_id
                }
            )

            d.refresh_from_db()

            d.process_error.append(exc.message)
            d.status = Data.STATUS_ERROR

            try:
                d.save(update_fields=['process_error', 'status'])
            except Exception:  # pylint: disable=broad-except
                pass
        except Exception:  # pylint: disable=broad-except
            logger.error(
                __(
                    "Error when saving Data object of process '{}' (handle_update):\n\n{}",
                    d.process.slug,
                    traceback.format_exc()
                ),
                extra={
                    'data_id': data_id
                }
            )

        if not internal_call:
            async_to_sync(self._send_reply)(obj, {ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK})

    def handle_finish(self, obj):
        """Handle an incoming ``Data`` finished processing request.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'finish',
                    'data_id': [id of the :class:`~resolwe.flow.models.Data` object
                               this command changes],
                    'process_rc': [exit status of the processing]
                    'spawn_processes': [optional; list of spawn dictionaries],
                    'exported_files_mapper': [if spawn_processes present]
                }
        """
        data_id = obj[ExecutorProtocol.DATA_ID]
        logger.debug(
            __("Finishing Data with id {} (handle_finish).", data_id),
            extra={
                'data_id': data_id,
                'packet': obj
            }
        )

        with transaction.atomic():
            # Spawn any new jobs in the request.
            spawned = False
            if ExecutorProtocol.FINISH_SPAWN_PROCESSES in obj:
                if is_testing():
                    # NOTE: This is a work-around for Django issue #10827
                    # (https://code.djangoproject.com/ticket/10827), same as in
                    # TestCaseHelpers._pre_setup(). Because the listener is running
                    # independently, it must clear the cache on its own.
                    ContentType.objects.clear_cache()

                spawned = True
                exported_files_mapper = obj[ExecutorProtocol.FINISH_EXPORTED_FILES]
                logger.debug(
                    __("Spawning new Data objects for Data with id {} (handle_finish).", data_id),
                    extra={
                        'data_id': data_id
                    }
                )

                try:
                    # This transaction is needed because we're running
                    # asynchronously with respect to the main Django code
                    # here; the manager can get nudged from elsewhere.
                    with transaction.atomic():
                        parent_data = Data.objects.get(pk=data_id)

                        # Spawn processes.
                        for d in obj[ExecutorProtocol.FINISH_SPAWN_PROCESSES]:
                            d['contributor'] = parent_data.contributor
                            d['process'] = Process.objects.filter(slug=d['process']).latest()
                            d['tags'] = parent_data.tags

                            for field_schema, fields in iterate_fields(d.get('input', {}), d['process'].input_schema):
                                type_ = field_schema['type']
                                name = field_schema['name']
                                value = fields[name]

                                if type_ == 'basic:file:':
                                    fields[name] = self.hydrate_spawned_files(
                                        exported_files_mapper, value, data_id
                                    )
                                elif type_ == 'list:basic:file:':
                                    fields[name] = [self.hydrate_spawned_files(exported_files_mapper, fn, data_id)
                                                    for fn in value]

                            with transaction.atomic():
                                d = Data.objects.create(**d)
                                DataDependency.objects.create(
                                    parent=parent_data,
                                    child=d,
                                    kind=DataDependency.KIND_SUBPROCESS,
                                )

                                # Copy permissions.
                                copy_permissions(parent_data, d)

                                # Entity is added to the collection only when it is
                                # created - when it only contains 1 Data object.
                                entities = Entity.objects.filter(data=d).annotate(num_data=Count('data')).filter(
                                    num_data=1)

                                # Copy collections.
                                for collection in parent_data.collection_set.all():
                                    collection.data.add(d)

                                    # Add entities to which data belongs to the collection.
                                    for entity in entities:
                                        entity.collections.add(collection)

                except Exception:  # pylint: disable=broad-except
                    logger.error(
                        __(
                            "Error while preparing spawned Data objects of process '{}' (handle_finish):\n\n{}",
                            parent_data.process.slug,
                            traceback.format_exc()
                        ),
                        extra={
                            'data_id': data_id
                        }
                    )

            # Data wrap up happens last, so that any triggered signals
            # already see the spawned children. What the children themselves
            # see is guaranteed by the transaction we're in.
            if ExecutorProtocol.FINISH_PROCESS_RC in obj:
                process_rc = obj[ExecutorProtocol.FINISH_PROCESS_RC]

                try:
                    d = Data.objects.get(pk=data_id)
                except Data.DoesNotExist:
                    logger.warning(
                        "Data object does not exist (handle_finish).",
                        extra={
                            'data_id': data_id,
                        }
                    )
                    async_to_sync(self._send_reply)(obj, {ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_ERROR})
                    return

                if process_rc == 0 and not d.status == Data.STATUS_ERROR:
                    changeset = {
                        'status': Data.STATUS_DONE,
                        'process_progress': 100,
                        'finished': now()
                    }
                else:
                    changeset = {
                        'status': Data.STATUS_ERROR,
                        'process_progress': 100,
                        'process_rc': process_rc,
                        'finished': now()
                    }
                obj[ExecutorProtocol.UPDATE_CHANGESET] = changeset
                self.handle_update(obj, internal_call=True)

        if not getattr(settings, 'FLOW_MANAGER_KEEP_DATA', False):
            # Purge worker is not running in test runner, so we should skip triggering it.
            if not is_testing():
                channel_layer = get_channel_layer()
                try:
                    async_to_sync(channel_layer.send)(
                        CHANNEL_PURGE_WORKER,
                        {
                            'type': TYPE_PURGE_RUN,
                            'location_id': d.location.id,
                            'verbosity': self._verbosity,
                        }
                    )
                except ChannelFull:
                    logger.warning(
                        "Cannot trigger purge because channel is full.",
                        extra={'data_id': data_id}
                    )

        # Notify the executor that we're done.
        async_to_sync(self._send_reply)(obj, {ExecutorProtocol.RESULT: ExecutorProtocol.RESULT_OK})

        # Now nudge the main manager to perform final cleanup. This is
        # needed even if there was no spawn baggage, since the manager
        # may need to know when executors have finished, to keep count
        # of them and manage synchronization.
        async_to_sync(consumer.send_event)({
            WorkerProtocol.COMMAND: WorkerProtocol.FINISH,
            WorkerProtocol.DATA_ID: data_id,
            WorkerProtocol.FINISH_SPAWNED: spawned,
            WorkerProtocol.FINISH_COMMUNICATE_EXTRA: {
                'executor': getattr(settings, 'FLOW_EXECUTOR', {}).get('NAME', 'resolwe.flow.executors.local'),
            },
        })

    def handle_abort(self, obj):
        """Handle an incoming ``Data`` abort processing request.

        .. IMPORTANT::

            This only makes manager's state consistent and doesn't
            affect Data object in any way. Any changes to the Data
            must be applied over ``handle_update`` method.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'abort',
                    'data_id': [id of the :class:`~resolwe.flow.models.Data` object
                               this command was triggered by],
                }
        """
        async_to_sync(consumer.send_event)({
            WorkerProtocol.COMMAND: WorkerProtocol.ABORT,
            WorkerProtocol.DATA_ID: obj[ExecutorProtocol.DATA_ID],
            WorkerProtocol.FINISH_COMMUNICATE_EXTRA: {
                'executor': getattr(settings, 'FLOW_EXECUTOR', {}).get('NAME', 'resolwe.flow.executors.local'),
            },
        })

    def handle_log(self, obj):
        """Handle an incoming log processing request.

        :param obj: The Channels message object. Command object format:

            .. code-block:: none

                {
                    'command': 'log',
                    'message': [log message]
                }
        """
        record_dict = json.loads(obj[ExecutorProtocol.LOG_MESSAGE])
        record_dict['msg'] = record_dict['msg']

        executors_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'executors')
        record_dict['pathname'] = os.path.join(executors_dir, record_dict['pathname'])

        logger.handle(logging.makeLogRecord(record_dict))

    def _make_stats(self):
        """Create a stats snapshot."""
        return {
            'load_avg': self.load_avg.to_dict(),
            'service_time': self.service_time.to_dict(),
        }

    async def push_stats(self):
        """Push current stats to Redis."""
        snapshot = self._make_stats()
        try:
            serialized = json.dumps(snapshot)
            await self._call_redis(aioredis.Redis.set, state.MANAGER_LISTENER_STATS, serialized)
            await self._call_redis(aioredis.Redis.expire, state.MANAGER_LISTENER_STATS, 3600)
        except TypeError:
            logger.error(__(
                "Listener can't serialize statistics:\n\n{}",
                traceback.format_exc()
            ))
        except aioredis.RedisError:
            logger.error(__(
                "Listener can't store updated statistics:\n\n{}",
                traceback.format_exc()
            ))

    def check_critical_load(self):
        """Check for critical load and log an error if necessary."""
        if self.load_avg.intervals['1m'].value > 1:
            if self.last_load_level == 1 and time.time() - self.last_load_log < 30:
                return
            self.last_load_log = time.time()
            self.last_load_level = 1
            logger.error(
                "Listener load limit exceeded, the system can't handle this!",
                extra=self._make_stats()
            )

        elif self.load_avg.intervals['1m'].value > 0.8:
            if self.last_load_level == 0.8 and time.time() - self.last_load_log < 30:
                return
            self.last_load_log = time.time()
            self.last_load_level = 0.8
            logger.warning(
                "Listener load approaching critical!",
                extra=self._make_stats()
            )

        else:
            self.last_load_log = -math.inf
            self.last_load_level = 0

    async def run(self):
        """Run the main listener run loop.

        Doesn't return until :meth:`terminate` is called.
        """
        logger.info(__(
            "Starting Resolwe listener on channel '{}'.",
            state.MANAGER_EXECUTOR_CHANNELS.queue
        ))
        while not self._should_stop:
            await self.push_stats()
            ret = await self._call_redis(aioredis.Redis.blpop, state.MANAGER_EXECUTOR_CHANNELS.queue, timeout=1)
            if ret is None:
                self.load_avg.add(0)
                continue
            remaining = await self._call_redis(aioredis.Redis.llen, state.MANAGER_EXECUTOR_CHANNELS.queue)
            self.load_avg.add(remaining + 1)
            self.check_critical_load()
            _, item = ret
            try:
                item = item.decode('utf-8')
                logger.debug(__("Got command from executor: {}", item))
                obj = json.loads(item)
            except json.JSONDecodeError:
                logger.error(
                    __("Undecodable command packet:\n\n{}"),
                    traceback.format_exc()
                )
                continue

            command = obj.get(ExecutorProtocol.COMMAND, None)
            if command is None:
                continue

            service_start = time.perf_counter()

            handler = getattr(self, 'handle_' + command, None)
            if handler:
                try:
                    with PrioritizedBatcher.global_instance():
                        await database_sync_to_async(handler)(obj)
                except Exception:  # pylint: disable=broad-except
                    logger.error(__(
                        "Executor command handling error:\n\n{}",
                        traceback.format_exc()
                    ))
            else:
                logger.error(
                    __("Unknown executor command '{}'.", command),
                    extra={'decoded_packet': obj}
                )

            # We do want to measure wall-clock time elapsed, because
            # system load will impact event handling performance. On
            # a lagging system, good internal performance is meaningless.
            service_end = time.perf_counter()
            self.service_time.update(service_end - service_start)
        logger.info(__(
            "Stopping Resolwe listener on channel '{}'.",
            state.MANAGER_EXECUTOR_CHANNELS.queue
        ))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-12-06 05:15
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import fernet_fields.fields
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('flow', '0009_make_size_mandatory'),
    ]

    operations = [
        migrations.CreateModel(
            name='Secret',
            fields=[
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True, db_index=True)),
                ('handle', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('value', fernet_fields.fields.EncryptedTextField()),
                ('metadata', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('expires', models.DateTimeField(db_index=True, null=True)),
                ('contributor', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



""".. Ignore pydocstyle D400.

==========
Statistics
==========

Various statistical utilities, used mostly for manager load tracking.
"""
import math
import time
from collections import deque, namedtuple


class NumberSeriesShape:
    """Helper class for computing characteristics for numerical data.

    Given a series of numerical data, the class will keep a record of
    the extremes seen, arithmetic mean and standard deviation.
    """

    def __init__(self):
        """Construct an instance of the class."""
        self.high = -math.inf
        self.low = math.inf
        self.mean = 0
        self.deviation = 0
        self.count = 0
        self._rolling_variance = 0

    def update(self, num):
        """Update metrics with the new number."""
        num = float(num)
        self.count += 1
        self.low = min(self.low, num)
        self.high = max(self.high, num)

        # Welford's online mean and variance algorithm.
        delta = num - self.mean
        self.mean = self.mean + delta / self.count
        delta2 = num - self.mean
        self._rolling_variance = self._rolling_variance + delta * delta2

        if self.count > 1:
            self.deviation = math.sqrt(self._rolling_variance / (self.count - 1))
        else:
            self.deviation = 0.0

    def to_dict(self):
        """Pack the stats computed into a dictionary."""
        return {
            'high': self.high,
            'low': self.low,
            'mean': self.mean,
            'count': self.count,
            'deviation': self.deviation,
        }


def _display_interval(i):
    """Convert a time interval into a human-readable string.

    :param i: The interval to convert, in seconds.
    """
    sigils = ["d", "h", "m", "s"]
    factors = [24 * 60 * 60, 60 * 60, 60, 1]
    remain = int(i)
    result = ""
    for fac, sig in zip(factors, sigils):
        if remain < fac:
            continue
        result += "{}{}".format(remain // fac, sig)
        remain = remain % fac
    return result


class SimpleLoadAvg:
    """Helper class for a sort of load average based on event times.

    Given a series of queue depth events, it will compute the average
    number of events for three different window lengths, emulating a
    form of 'load average'. The calculation itself is modelled after the
    Linux scheduler, with a 5-second sampling rate. Because we don't get
    consistent (time-wise) samples, the sample taken is the average of a
    simple moving window for the last 5 seconds; this is to avoid
    numerical errors if actual time deltas were used to compute the
    scaled decay.
    """

    class _Interval:
        """Convenience class containing bookkeeping for an interval."""

        Point = namedtuple('Point', ['time', 'count'])

        def __init__(self, interval):
            """Construct an instance of the class.

            :param interval: The interval to represent.
            """
            self.series = deque()
            self.sampling_window = 5
            self.interval = interval
            self.display = _display_interval(interval)
            self.last_bound = None
            self.decay = 1 / math.exp(self.sampling_window / self.interval)
            self.value = 0

        def push(self, count, timestamp):
            """Add a new data point to the window."""
            sample = self.Point(timestamp, count)

            # Check if we're still filling up the current window.
            if self.last_bound is None:
                self.last_bound = sample.time
                self.series.appendleft(sample)
                return
            if not self.series or sample.time - self.last_bound < self.sampling_window:
                self.series.appendleft(sample)
                return

            # If the window is full, get the average for the window size we want.
            # This will lead to some jitter if the actual time difference
            # between the latest sample and the earliest one isn't spot on,
            # but that's why this class isn't ComplexLoadAvg.
            avg = 0
            for sample in self.series:
                avg += sample.count
            avg /= len(self.series)

            # Clear storage for the next window.
            self.last_bound = timestamp
            self.series.clear()

            # The actual EMA computation.
            self.value = self.decay * self.value + (1 - self.decay) * avg

    def __init__(self, intervals):
        """Construct an instance of the class.

        :param interval: A list of interval lengths, in seconds.
        """
        self.last_data = -math.inf
        self.intervals = {i: SimpleLoadAvg._Interval(i) for i in intervals}
        for meta in list(self.intervals.values()):
            self.intervals[meta.display] = meta

    def add(self, count, timestamp=None):
        """Add a value at the specified time to the series.

        :param count: The number of work items ready at the specified
            time.
        :param timestamp: The timestamp to add. Defaults to None,
            meaning current time. It should be strictly greater (newer)
            than the last added timestamp.
        """
        if timestamp is None:
            timestamp = time.time()
        if self.last_data >= timestamp:
            raise ValueError("Time {} >= {} in load average calculation".format(self.last_data, timestamp))
        self.last_data = timestamp

        for meta in self.intervals.values():
            meta.push(count, timestamp)

    def to_dict(self):
        """Pack the load averages into a nicely-keyed dictionary."""
        result = {}
        for meta in self.intervals.values():
            result[meta.display] = meta.value
        return result



""".. Ignore pydocstyle D400.

===================
Resolwe Test Runner
===================

"""
import asyncio
import contextlib
import logging
import os
import re
import shutil
import subprocess
import sys

import mock
import yaml
from channels.db import database_sync_to_async

from django.conf import settings
from django.core.management.base import CommandError
from django.test import override_settings
from django.test.runner import DiscoverRunner, ParallelTestSuite, RemoteTestRunner
from django.utils.crypto import get_random_string

# Make sure we already have the patched FLOW_* available here; otherwise
# resolwe.test.testcases.TransactionTestCase will override them with the module above,
# negating anything we do here with Django's override_settings.
import resolwe.test.testcases.setting_overrides as resolwe_settings
from resolwe.flow.finders import get_finders
from resolwe.flow.managers import manager, state
from resolwe.flow.managers.listener import ExecutorListener
from resolwe.test.utils import generate_process_tag

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

SPAWN_PROCESS_REGEX = re.compile(r'run\s+\{.*?["\']process["\']\s*:\s*["\'](.+?)["\'].*?\}')

TESTING_CONTEXT = {
    'is_testing': False,
}


class TestingContext:
    """Context manager which maintains current testing status."""

    def __enter__(self):
        """Enter testing context."""
        TESTING_CONTEXT['is_testing'] = True

    def __exit__(self, *args, **kwargs):
        """Exit testing context."""
        TESTING_CONTEXT['is_testing'] = False

        # Propagate exceptions.
        return False


class AtScopeExit:
    """Utility class for calling a function once a context exits."""

    def __init__(self, call, *args, **kwargs):
        """Construct a context manager and save arguments.

        :param call: The callable to call on exit.
        :param args: Positional arguments for the callable.
        :param kwargs: Keyword arguments for the callable.
        """
        self.call = call
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        """Enter the ``with`` context."""
        return self

    def __exit__(self, *args, **kwargs):
        """Exit the context and call the saved callable."""
        self.call(*self.args, **self.kwargs)
        return False


def _manager_setup():
    """Execute setup operations common to serial and parallel testing.

    This mostly means state cleanup, such as resetting database
    connections and clearing the shared state.
    """
    if TESTING_CONTEXT.get('manager_reset', False):
        return
    TESTING_CONTEXT['manager_reset'] = True
    state.update_constants()
    manager.reset()


def _sequence_paths(paths):
    """Extend the last components of the given paths with a number.

    The method finds the lowest number such that all given paths, when
    extended by it, are unique and can be created. The paths are then
    also created.

    :param paths: The list of paths to be extended and created.
    :return: The list of created paths.
    """
    seq = 0
    while True:
        # Note for parallel execution: infinite zigzagging ladders are
        # not possible, because the directories are always created in
        # the same order. The problem would be if process A succeeded
        # in creating data/test_1, but process B would beat it to
        # upload/test_1 (so that both would roll back and continue
        # with _2, etc.). For B to succeed in creating upload/test_1,
        # it must have already succeeded in creating data/test_1,
        # meaning A could not possibly have succeeded with data/test_1.
        seq += 1
        created = []

        for base_path in paths:
            path = os.path.join(base_path, 'test_{}'.format(seq))
            try:
                os.makedirs(path)
                created.append(path)
            except OSError:
                break

        if len(created) == len(paths):
            return created

        # If they're not equal, we failed and need to roll back;
        # errors are entirely irrelevant here, removal is purely
        # best effort.
        for path in created:
            try:
                os.rmdir(path)
            except:  # noqa pylint: disable=bare-except
                pass


def _create_test_dirs():
    """Create all the testing directories."""
    if 'test_paths' in TESTING_CONTEXT:
        return TESTING_CONTEXT['test_paths']
    items = ['DATA_DIR', 'UPLOAD_DIR', 'RUNTIME_DIR']
    paths = _sequence_paths([resolwe_settings.FLOW_EXECUTOR_SETTINGS[i] for i in items])
    for item, path in zip(items, paths):
        resolwe_settings.FLOW_EXECUTOR_SETTINGS[item] = path
    TESTING_CONTEXT['test_paths'] = paths
    return paths


def _prepare_settings():
    """Prepare and apply settings overrides needed for testing."""
    # Override container name prefix setting.
    resolwe_settings.FLOW_EXECUTOR_SETTINGS['CONTAINER_NAME_PREFIX'] = '{}_{}_{}'.format(
        resolwe_settings.FLOW_EXECUTOR_SETTINGS.get('CONTAINER_NAME_PREFIX', 'resolwe'),
        # NOTE: This is necessary to avoid container name clashes when tests are run from
        # different Resolwe code bases on the same system (e.g. on a CI server).
        get_random_string(length=6),
        os.path.basename(resolwe_settings.FLOW_EXECUTOR_SETTINGS['DATA_DIR'])
    )

    return override_settings(
        CELERY_ALWAYS_EAGER=True,
        FLOW_EXECUTOR=resolwe_settings.FLOW_EXECUTOR_SETTINGS,
        FLOW_MANAGER=resolwe_settings.FLOW_MANAGER_SETTINGS,
    )


def _custom_worker_init(django_init_worker):
    """Wrap the original worker init to also start the manager."""
    def _init_worker(*args, **kwargs):
        """Initialize a :class:`multiprocessing.Pool` worker.

        Call the Django's ``ParallelTestSuite.init_worker`` and then
        also start the manager infrastructure.
        """
        result = django_init_worker(*args, **kwargs)

        # Further patch channel names and the like with our current pid,
        # so that parallel managers and executors don't clash on the
        # same channels and directories.
        resolwe_settings.FLOW_MANAGER_SETTINGS['REDIS_PREFIX'] += '-parallel-pid{}'.format(os.getpid())

        return result
    return _init_worker


def _run_in_event_loop(coro, *args, **kwargs):
    """Run a coroutine in a runloop call.

    This is needed as the top level call into Resolwe Manager-using
    tests. An event loop is started so that it can be used within the
    call tree.

    :param coro: The coroutine to run with an underlying event loop. All
        other arguments given to this function are forwarded to it.
    """
    asyncio.get_event_loop().close()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    task = asyncio.ensure_future(coro(*args, **kwargs), loop=loop)
    loop.run_until_complete(task)
    loop.close()
    return task.result()


async def _run_on_infrastructure(meth, *args, **kwargs):
    """Start the Manager infrastructure and call the given callable.

    The method given is run through a serializing wrapper, so that
    Django database accesses are correct.

    :param meth: The callable to run on the infrastructure. All other
        arguments are forwarded to it.
    """
    with TestingContext():
        _create_test_dirs()
        with _prepare_settings():
            await database_sync_to_async(_manager_setup)()
            with AtScopeExit(manager.state.destroy_channels):
                redis_params = getattr(settings, 'FLOW_MANAGER', {}).get('REDIS_CONNECTION', {})
                listener = ExecutorListener(redis_params=redis_params)
                await listener.clear_queue()
                async with listener:
                    try:
                        with override_settings(FLOW_MANAGER_SYNC_AUTO_CALLS=True):
                            result = await database_sync_to_async(meth)(*args, **kwargs)
                        return result
                    finally:
                        listener.terminate()


def _run_manager(meth, *args, **kwargs):
    """Start the Manager properly and nest the given callable in it.

    :param meth: The callable to start the environment for; all other
        arguments are forwarded to it.
    """
    return _run_in_event_loop(_run_on_infrastructure, meth, *args, **kwargs)


class CustomRemoteRunner(RemoteTestRunner):
    """Standard Django remote runner with a custom run method."""

    def run(self, *args, **kwargs):
        """Run the superclass method with an underlying event loop."""
        # NOTE: An instance of this class is constructed for _each_ batch
        # of tests, not just once per pool worker, so some care needs to
        # be taken to avoid incremental corruption, such as in
        # _create_test_dirs().
        return _run_manager(super().run, *args, **kwargs)


class CustomParallelTestSuite(ParallelTestSuite):
    """Standard parallel suite with a custom worker initializer."""

    init_worker = _custom_worker_init(ParallelTestSuite.init_worker)
    runner_class = CustomRemoteRunner


class ResolweRunner(DiscoverRunner):
    """Resolwe test runner."""

    parallel_test_suite = CustomParallelTestSuite

    def __init__(self, *args, **kwargs):
        """Initialize test runner."""
        self.only_changes_to = kwargs.pop('only_changes_to', None)
        self.changes_file_types = kwargs.pop('changes_file_types', None)

        # Handle implication first, meanings get inverted later.
        self.keep_data = False
        if 'no_mock_purge' in kwargs:
            self.keep_data = True

        # mock_purge used to be the default, so keep it that way;
        # this means the command line option does the opposite
        # (disables it), so the boolean must be inverted here.
        self.mock_purge = not kwargs.pop('no_mock_purge', False)

        self.keep_data = kwargs.pop('keep_data', self.keep_data)

        super().__init__(*args, **kwargs)

    @classmethod
    def add_arguments(cls, parser):
        """Add command-line arguments.

        :param parser: Argument parser instance
        """
        super().add_arguments(parser)

        parser.add_argument(
            '--only-changes-to', dest='only_changes_to',
            help="Only test changes against given Git commit reference"
        )

        parser.add_argument(
            '--changes-file-types', dest='changes_file_types',
            help="File which describes what kind of changes are available"
        )

        parser.add_argument(
            '--keep-data', dest='keep_data', action='store_true',
            help="Prevent test cases from cleaning up after execution"
        )

        parser.add_argument(
            '--no-mock-purge', dest='no_mock_purge', action='store_true',
            help="Do not mock purging functions (implies --keep-data)"
        )

    def build_suite(self, *args, **kwargs):
        """Build test suite."""
        suite = super().build_suite(*args, **kwargs)
        # Build suite first constructs the parallel suite and then may reduce self.parallel,
        # while keeping suite.processes unchanged. We need to propagate the change here to
        # avoid spawning more processes than there are databases.
        suite.processes = self.parallel

        # Augment all test cases with manager state validation logic.
        def validate_manager_state(case, teardown):
            """Decorate test case with manager state validation."""
            def wrapper(*args, **kwargs):
                """Validate manager state on teardown."""
                if manager.sync_counter.value != 0:
                    case.fail(
                        'Test has outstanding manager processes. Ensure that all processes have '
                        'completed or that you have reset the state manually in case you have '
                        'bypassed the regular manager flow in any way.\n'
                        '\n'
                        'Synchronization count: {value} (should be 0)\n'
                        ''.format(
                            value=manager.sync_counter.value,
                        )
                    )

                teardown(*args, **kwargs)

            return wrapper

        for case in suite:
            case.tearDown = validate_manager_state(case, case.tearDown)

        return suite

    def run_suite(self, suite, **kwargs):
        """Run the test suite with manager workers in the background."""
        # Due to the way the app modules are imported, there's no way to
        # statically override settings with TEST overrides before e.g.
        # resolwe.flow.managers.manager is loaded somewhere in the code
        # (the most problematic files are signals.py and
        # resolwe.test.testcases.process); the only realistic mechanism
        # is to override later and call some sort of commit method in
        # the manager.

        keep_data_override = override_settings(FLOW_MANAGER_KEEP_DATA=self.keep_data)
        keep_data_override.__enter__()

        if self.keep_data and self.mock_purge:
            purge_mock_os = mock.patch('resolwe.flow.utils.purge.os', wraps=os).start()
            purge_mock_os.remove = mock.MagicMock()

            purge_mock_shutil = mock.patch('resolwe.flow.utils.purge.shutil', wraps=shutil).start()
            purge_mock_shutil.rmtree = mock.MagicMock()

        if self.parallel > 1:
            return super().run_suite(suite, **kwargs)

        return _run_manager(super().run_suite, suite, **kwargs)

    def run_tests(self, test_labels, **kwargs):
        """Run tests.

        :param test_labels: Labels of tests to run
        """
        if self.only_changes_to:
            # Check if there are changed files. We need to be able to switch between branches so we
            # can correctly detect changes.
            repo_status = self._git('status', '--porcelain', '--untracked-files=no')
            if repo_status:
                print("ERROR: Git repository is not clean. Running tests with --only-changes-to", file=sys.stderr)
                print("       requires that the current Git repository is clean.", file=sys.stderr)
                return False

            print("Detecting changed files between {} and HEAD.".format(self.only_changes_to))
            changed_files = self._git('diff', '--name-only', self.only_changes_to, 'HEAD')
            changed_files = changed_files.strip().split('\n')
            changed_files = [file for file in changed_files if file.strip()]

            top_level_path = self._git('rev-parse', '--show-toplevel')
            top_level_path = top_level_path.strip()

            # Process changed files to discover what they are.
            changed_files, tags, tests, full_suite = self.process_changed_files(changed_files, top_level_path)
            print("Changed files:")
            for filename, file_type in changed_files:
                print("  {} ({})".format(filename, file_type))

            if not changed_files:
                print("  none")
                print("No files have been changed, assuming target is HEAD, running full suite.")
            elif full_suite:
                print("Non-test code or unknown files have been modified, running full test suite.")
            else:
                # Run specific tests/tags.
                print("Running with following partial tags: {}".format(', '.join(tags)))
                print("Running with following partial tests: {}".format(', '.join(tests)))

                failed_tests = 0

                # First run with specific tags. Since run_tests may modify self.parallel, we need to store
                # it here and restore it later if we also run with specific test labels.
                parallel = self.parallel
                if tags:
                    self.tags = tags
                    failed_tests += super().run_tests(test_labels, **kwargs)

                # Then run with specific test labels.
                if tests:
                    self.parallel = parallel
                    self.tags = set()
                    failed_tests += super().run_tests(tests, **kwargs)

                return failed_tests

        return super().run_tests(test_labels, **kwargs)

    def _git(self, *args):
        """Helper to run Git command."""
        try:
            return subprocess.check_output(['git'] + list(args)).decode('utf8').strip()
        except subprocess.CalledProcessError:
            raise CommandError("Git command failed.")

    @contextlib.contextmanager
    def git_branch(self, branch):
        """Temporarily switch to a different Git branch."""
        current_branch = self._git('rev-parse', '--abbrev-ref', 'HEAD')
        if current_branch == 'HEAD':
            # Detached HEAD state, we need to get the actual commit.
            current_branch = self._git('rev-parse', 'HEAD')

        if current_branch != branch:
            self._git('checkout', branch)

        try:
            yield
        finally:
            if current_branch != branch:
                self._git('checkout', current_branch)

    def process_changed_files(self, changed_files, top_level_path):
        """Process changed files based on specified patterns.

        :param changed_files: A list of changed file pats, relative to top-level path
        :param top_level_path: Absolute path to top-level project directory
        :return: Tuple (changed_files, tags, tests, full_suite)
        """
        result = []
        processes = []
        tests = []
        full_suite = False
        types = []

        if self.changes_file_types:
            # Parse file types metadata.
            try:
                with open(self.changes_file_types, 'r') as definition_file:
                    types = yaml.load(definition_file)
            except (OSError, ValueError):
                raise CommandError("Failed loading or parsing file types metadata.")
        else:
            print("WARNING: Treating all files as unknown because --changes-file-types option not specified.",
                  file=sys.stderr)

        for filename in changed_files:
            # Match file type.
            file_type = 'unknown'
            file_type_name = 'unknown'

            for definition in types:
                if re.search(definition['match'], filename):
                    file_type = definition['type']
                    file_type_name = definition.get('name', file_type)
                    break

            result.append((filename, file_type_name))
            if file_type in ('unknown', 'force_run'):
                full_suite = True
            elif file_type == 'ignore':
                # Ignore
                pass
            elif file_type == 'process':
                # Resolve process tag.
                processes.append(os.path.join(top_level_path, filename))
            elif file_type == 'test':
                # Generate test name.
                tests.append(re.sub(r'\.py$', '', filename).replace(os.path.sep, '.'))
            else:
                raise CommandError("Unsupported file type: {}".format(file_type))

        # Resolve tags.
        tags = self.resolve_process_tags(processes)

        return result, tags, tests, full_suite

    def find_schemas(self, schema_path):
        """Find process schemas.

        :param schema_path: Path where to look for process schemas
        :return: Found schemas
        """
        schema_matches = []

        for root, _, files in os.walk(schema_path):
            for schema_file in [os.path.join(root, fn) for fn in files]:

                if not schema_file.lower().endswith(('.yml', '.yaml')):
                    continue

                with open(schema_file) as fn:
                    schemas = yaml.load(fn)

                if not schemas:
                    print("WARNING: Could not read YAML file {}".format(schema_file), file=sys.stderr)
                    continue

                for schema in schemas:
                    schema_matches.append(schema)

        return schema_matches

    def find_dependencies(self, schemas):
        """Compute process dependencies.

        :param schemas: A list of all discovered process schemas
        :return: Process dependency dictionary
        """
        dependencies = {}

        for schema in schemas:
            slug = schema['slug']
            run = schema.get('run', {})
            program = run.get('program', None)
            language = run.get('language', None)

            if language == 'workflow':
                for step in program:
                    dependencies.setdefault(step['run'], set()).add(slug)
            elif language == 'bash':
                # Process re-spawn instructions to discover dependencies.
                matches = SPAWN_PROCESS_REGEX.findall(program)
                if matches:
                    for match in matches:
                        dependencies.setdefault(match, set()).add(slug)

        return dependencies

    def resolve_process_tags(self, files):
        """Resolve process tags.

        :param files: List of changed process files
        :return: Test tags that need to be run
        """
        processes_paths = []
        for finder in get_finders():
            processes_paths.extend(finder.find_processes())

        process_schemas = []
        for proc_path in processes_paths:
            process_schemas.extend(self.find_schemas(proc_path))

        # Switch to source branch and get all the schemas from there as well, since some schemas
        # might have been removed.
        with self.git_branch(self.only_changes_to):
            for proc_path in processes_paths:
                process_schemas.extend(self.find_schemas(proc_path))

        dependencies = self.find_dependencies(process_schemas)
        processes = set()

        def load_process_slugs(filename):
            """Add all process slugs from specified file."""
            with open(filename, 'r') as process_file:
                data = yaml.load(process_file)

                for process in data:
                    # Add all process slugs.
                    processes.add(process['slug'])

        for filename in files:
            try:
                load_process_slugs(filename)
            except FileNotFoundError:
                # File was removed, so we will handle it below when we check the original branch.
                pass

        # Switch to source branch and check modified files there as well.
        with self.git_branch(self.only_changes_to):
            for filename in files:
                try:
                    load_process_slugs(filename)
                except FileNotFoundError:
                    # File was added, so it has already been handled.
                    pass

        # Add all dependencies.
        dep_processes = set()
        while processes:
            process = processes.pop()
            if process in dep_processes:
                continue

            dep_processes.add(process)
            processes.update(dependencies.get(process, set()))

        tags = set()
        for process in dep_processes:
            tags.add(generate_process_tag(process))

        return tags


def is_testing():
    """Return current testing status."""
    return TESTING_CONTEXT['is_testing']



""".. Ignore pydocstyle D400.

===============
Flow Management
===============

.. automodule:: resolwe.flow.management.commands.purge
    :members:

.. automodule:: resolwe.flow.management.commands.register
    :members:

"""



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-09-06 15:23
from __future__ import unicode_literals

import django.contrib.postgres.fields.citext
from django.contrib.postgres.operations import CITextExtension
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0018_process_is_active'),
    ]

    operations = [
        CITextExtension(),
        migrations.AlterField(
            model_name='relation',
            name='category',
            field=django.contrib.postgres.fields.citext.CICharField(max_length=100),
        ),
    ]



from resolwe.process import *


class PythonProcessJson(Process):
    """This is a process description."""
    slug = 'test-python-process-json'
    name = "Python Process that uses JSON field"
    process_type = 'data:python'
    version = '0.1.2'
    requirements = {
        'executor': {
            'docker': {
                'image': 'resolwe/base:ubuntu-18.04',
            }
        }
    }

    class Input:
        """Input fields."""
        data = DataField('test', label="My input data")

    def run(self, inputs, outputs):
        print('Input data:', inputs.data)
        print('Input data:', inputs.data.storage)



"""Resolwe storage serializer."""
from resolwe.flow.models import Storage
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer


class StorageSerializer(ResolweBaseSerializer):
    """Serializer for Storage objects."""

    json = ProjectableJSONField()

    class Meta:
        """StorageSerializer Meta options."""

        model = Storage
        read_only_fields = (
            'created',
            'id',
            'modified',
        )
        update_protected_fields = (
            'contributor',
        )
        fields = read_only_fields + update_protected_fields + (
            'data',
            'json',
            'name',
            'slug',
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-12 08:21
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0019_relation_type_cicharfield'),
    ]

    operations = [
        migrations.AlterField(
            model_name='process',
            name='category',
            field=models.CharField(default='Other:', max_length=200, validators=[django.core.validators.RegexValidator(code='invalid_category', message='Category may be alphanumerics separated by colon', regex='^([a-zA-Z0-9]+[:\\-])*[a-zA-Z0-9]+:$')]),
        ),
    ]



"""Protocol constants used by the manager and executors."""


class WorkerProtocol:
    """Constants used by the manager workers."""

    COMMAND = 'command'
    DATA_ID = 'data_id'

    COMMUNICATE = 'communicate'
    COMMUNICATE_SETTINGS = 'settings_override'
    COMMUNICATE_EXTRA = 'kwargs'

    FINISH = 'finish_data'
    FINISH_SPAWNED = 'spawned'
    FINISH_COMMUNICATE_EXTRA = 'communicate_kwargs'

    ABORT = 'abort_data'


class ExecutorProtocol:
    """Constants used by the executor<->listener protocol."""

    COMMAND = 'command'
    DATA_ID = 'data_id'

    UPDATE = 'update'
    UPDATE_CHANGESET = 'changeset'

    FINISH = 'finish'
    FINISH_PROCESS_RC = 'process_rc'
    FINISH_SPAWN_PROCESSES = 'spawn_processes'
    FINISH_EXPORTED_FILES = 'exported_files_mapper'

    ABORT = 'abort'

    RESULT = 'result'
    RESULT_OK = 'OK'
    RESULT_ERROR = 'ER'

    LOG = 'log'
    LOG_MESSAGE = 'message'


class ExecutorFiles:
    """Various files used by the executor."""

    FILE_LIST_KEY = 'serialized_files'

    EXECUTOR_SETTINGS = 'settings.json'
    DJANGO_SETTINGS = 'django_settings.json'
    DATA = 'data.json'
    DATA_LOCATION = 'data_location.json'
    DATA_META = 'data_meta.json'
    PROCESS = 'process.json'
    PROCESS_META = 'process_meta.json'

    PROCESS_SCRIPT = 'process_script.sh'

    SECRETS_DIR = 'secrets'



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-25 08:51
from __future__ import unicode_literals

from django.db import migrations

from resolwe.flow.utils import iterate_fields


def update_dependency_kinds(apps, schema_editor):
    """Update historical dependency kinds as they may be wrong."""
    DataDependency = apps.get_model('flow', 'DataDependency')
    for dependency in DataDependency.objects.all():
        # Assume dependency is of subprocess kind.
        dependency.kind = 'subprocess'

        # Check child inputs to determine if this is an IO dependency.
        child = dependency.child
        parent = dependency.parent

        for field_schema, fields in iterate_fields(child.input, child.process.input_schema):
            name = field_schema['name']
            value = fields[name]

            if field_schema.get('type', '').startswith('data:'):
                if value == parent.pk:
                    dependency.kind = 'io'
                    break
            elif field_schema.get('type', '').startswith('list:data:'):
                for data in value:
                    if value == parent.pk:
                        dependency.kind = 'io'
                        break

        dependency.save()


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0004_data_dependency_2'),
    ]

    operations = [
        migrations.RunPython(update_dependency_kinds)
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-03-01 02:25
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0030_add_data_duplicate'),
    ]

    operations = [
        migrations.AddField(
            model_name='entity',
            name='duplicated',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# pylint: disable=missing-docstring
from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.db.utils import DataError
from django.test import TestCase, override_settings

from resolwe.flow.exceptions import SlugError

# Workaround for false positive warnings in pylint.
# TODO: Can be removed with Django 2.0 or when pylint issue is fixed:
#       https://github.com/PyCQA/pylint/issues/1653
# pylint: disable=deprecated-method

CUSTOM_SETTINGS = {
    'INSTALLED_APPS': settings.INSTALLED_APPS + ('resolwe.flow.tests.fields_test_app',),
}


@override_settings(**CUSTOM_SETTINGS)
class FieldsTest(TestCase):

    def setUp(self):
        super().setUp()

        apps.clear_cache()
        call_command('migrate', verbosity=0, interactive=False, load_initial_data=False)
        # NOTE: Models must be imported after migrations are run.

    def test_populate_from_name(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name='Test object')
        self.assertEqual(obj.slug, 'test-object')

        obj = TestModel.objects.create(name='Test object')
        self.assertEqual(obj.slug, 'test-object-2')

        # Base of the slug shouldn't be changed even if it finishes with number.
        obj = TestModel.objects.create(name='Test object 2')
        self.assertEqual(obj.slug, 'test-object-2-2')

    def test_predefined_slug(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name='Test object')

        # It is ok to add a sequence to slug if it is not explicitly defined.
        TestModel.objects.create(name='Test object')

        with self.assertRaisesRegex(SlugError, 'is already taken'):
            TestModel.objects.create(name='Test object', slug=obj.slug)

    def test_predefined_long_slugs(self):
        from .fields_test_app.models import TestModel

        slug_max_length = TestModel._meta.get_field('slug').max_length  # pylint: disable=protected-access
        max_slug = 'x' * slug_max_length

        # Predefined slug shouldn't be trimmed
        obj = TestModel.objects.create(name='Test object', slug=max_slug)
        self.assertEqual(obj.slug, max_slug)

        with self.assertRaisesRegex(DataError, 'value too long'):
            TestModel.objects.create(name='Test object', slug=max_slug + 'y')

    def test_generate_only_on_create(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name='Test object')
        obj.save()  # Second save shouldn't add sequence number to slug.
        self.assertEqual(obj.slug, 'test-object')

        obj.name = 'Different name'
        obj.save()
        self.assertEqual(obj.slug, 'test-object')

    def test_unique_with(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name='Test object', version='1.0.0')
        self.assertEqual(obj.slug, 'test-object')

        obj = TestModel.objects.create(name='Test object', version='2.0.0')
        self.assertEqual(obj.slug, 'test-object')

        obj.version = '1.0.0'
        with self.assertRaisesRegex(SlugError, 'is already taken'):
            obj.save()

    def test_long_sequence(self):
        from .fields_test_app.models import TestModel

        TestModel.objects.create(name='Test object', slug='test-object')
        TestModel.objects.create(name='Test object', slug='test-object-999999999')

        with self.assertRaisesRegex(SlugError, 'slug sequence too long'):
            TestModel.objects.create(name='Test object')

    def test_empty_name(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create()
        self.assertEqual(obj.slug, 'testmodel')

        # Empty slugified name.
        obj = TestModel.objects.create(name='?')
        self.assertEqual(obj.slug, 'testmodel-2')



# pylint: disable=missing-docstring
from rest_framework import status

from resolwe.flow.models import Process
from resolwe.flow.views import ProcessViewSet
from resolwe.test import ResolweAPITestCase


class ProcessTestCase(ResolweAPITestCase):
    fixtures = ['users.yaml', 'processes.yaml', 'data.yaml', 'collections.yaml', 'permissions.yaml']

    def setUp(self):
        self.process1 = Process.objects.get(pk=1)

        self.post_data = {
            'slug': 'new-process',
            'name': 'New process',
            'type': 'data:test:process:',
            'input_schema': [{'name': 'test_field'}],
            'run': {'bash': 'echo $PATH'},
        }

        self.resource_name = 'process'
        self.viewset = ProcessViewSet

        super().setUp()

    def test_get_list(self):
        resp = self._get_list(self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_get_list_public_user(self):
        resp = self._get_list()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 1)

    def test_get_list_admin(self):
        resp = self._get_list(self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_post(self):
        resp = self._post(self.post_data, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

    def test_patch(self):
        resp = self._patch(1, {'name': 'Hacked process'}, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_delete(self):
        resp = self._delete(1, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)



# pylint: disable=missing-docstring,too-many-lines
from mock import MagicMock, patch

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError

from resolwe.flow.models import Collection, Data, DataLocation, DescriptorSchema, Entity, Process
from resolwe.flow.models.utils import validate_schema
from resolwe.test import TestCase

# Workaround for false positive warnings in pylint.
# TODO: Can be removed with Django 2.0 or when pylint issue is fixed:
#       https://github.com/PyCQA/pylint/issues/1653
# pylint: disable=deprecated-method


class ValidationTest(TestCase):

    def setUp(self):
        super().setUp()

        self.user = get_user_model().objects.create(username='test_user')

    def test_validating_data_object(self):
        """Diferent validations are performed depending on status"""
        proc = Process.objects.create(
            name='Test process',
            contributor=self.user,
            input_schema=[
                {'name': 'value', 'type': 'basic:integer:', 'required': True}
            ],
            output_schema=[
                {'name': 'result', 'type': 'basic:string:', 'required': True}
            ]
        )

        data = {
            'name': 'Test data',
            'contributor': self.user,
            'process': proc,
        }

        with self.assertRaisesRegex(ValidationError, '"value" not given'):
            Data.objects.create(input={}, **data)

        with self.assertRaisesRegex(ValidationError, 'Required fields .* not given'):
            Data.objects.create(input={}, **data)

        d = Data.objects.create(input={'value': 42}, **data)

        d.status = Data.STATUS_DONE
        with self.assertRaisesRegex(ValidationError, '"result" not given'):
            d.save()

        d.output = {'result': 'forty-two'}
        d.save()

    def test_validate_data_descriptor(self):
        proc = Process.objects.create(name='Test process', contributor=self.user)
        descriptor_schema = DescriptorSchema.objects.create(
            name='Test descriptor schema',
            contributor=self.user,
            schema=[
                {'name': 'description', 'type': 'basic:string:', 'required': True}
            ]
        )

        data = Data.objects.create(
            name='Test descriptor',
            contributor=self.user,
            process=proc,
            descriptor={},
            descriptor_schema=descriptor_schema
        )
        self.assertEqual(data.descriptor_dirty, True)

        data.descriptor = {'description': 'some value'}
        data.save()
        self.assertEqual(data.descriptor_dirty, False)

        data.descriptor = {}
        data.save()
        self.assertEqual(data.descriptor_dirty, True)

        data.descriptor = {'description': 42}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            data.save()

    def test_validate_collection_descriptor(self):  # pylint: disable=invalid-name
        descriptor_schema = DescriptorSchema.objects.create(
            name='Test descriptor schema',
            contributor=self.user,
            schema=[
                {'name': 'description', 'type': 'basic:string:', 'required': True}
            ]
        )

        collection = Collection.objects.create(
            name='Test descriptor',
            contributor=self.user,
            descriptor_schema=descriptor_schema
        )
        self.assertEqual(collection.descriptor_dirty, True)

        collection.descriptor = {'description': 'some value'}
        collection.save()
        self.assertEqual(collection.descriptor_dirty, False)

        collection.descriptor = {}
        collection.save()
        self.assertEqual(collection.descriptor_dirty, True)

        collection.descriptor = {'description': 42}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            collection.save()

    def test_validate_entity_descriptor(self):
        descriptor_schema = DescriptorSchema.objects.create(
            name='Test descriptor schema',
            contributor=self.user,
            schema=[
                {'name': 'description', 'type': 'basic:string:', 'required': True}
            ]
        )

        entity = Entity.objects.create(
            name='Test descriptor',
            contributor=self.user,
            descriptor_schema=descriptor_schema
        )
        self.assertEqual(entity.descriptor_dirty, True)

        entity.descriptor = {'description': 'some value'}
        entity.save()
        self.assertEqual(entity.descriptor_dirty, False)

        entity.descriptor = {}
        entity.save()
        self.assertEqual(entity.descriptor_dirty, True)

        entity.descriptor = {'description': 42}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            entity.save()

    def test_referenced_storage(self):
        proc = Process.objects.create(
            name='Test process',
            contributor=self.user,
            output_schema=[
                {'name': 'big_result', 'type': 'basic:json:', 'required': True}
            ]
        )

        data = {
            'name': 'Test data',
            'contributor': self.user,
            'process': proc,
        }

        d = Data.objects.create(**data)
        d.status = Data.STATUS_DONE

        # `Data` object with referenced non-existing `Storage`
        d.output = {'big_result': 245}
        with self.assertRaisesRegex(ValidationError, '`Storage` object does not exist'):
            d.save()

        d.storages.create(
            name="storage",
            contributor=self.user,
            json={'value': 42}
        )
        self.assertEqual(d.storages.count(), 1)
        d.output = {'big_result': d.storages.first().id}
        d.save()

    def test_referenced_data(self):
        proc1 = Process.objects.create(
            name='Referenced process',
            contributor=self.user,
            type='data:referenced:object:'
        )
        proc2 = Process.objects.create(
            name='Test process',
            contributor=self.user,
            input_schema=[
                {'name': 'data_object', 'type': 'data:referenced:object:'}
            ]
        )
        d = Data.objects.create(
            name='Referenced object',
            contributor=self.user,
            process=proc1
        )

        data = {
            'name': 'Test data',
            'contributor': self.user,
            'process': proc2,
            'input': {'data_object': d.pk}
        }

        Data.objects.create(**data)

        # less specific type
        proc2.input_schema = [
            {'name': 'data_object', 'type': 'data:referenced:'}
        ]
        Data.objects.create(**data)

        # wrong type
        proc2.input_schema = [
            {'name': 'data_object', 'type': 'data:wrong:type:'}
        ]
        with self.assertRaisesRegex(ValidationError, 'Data object of type .* is required'):
            Data.objects.create(**data)

        # non-existing `Data` object
        data['input'] = {'data_object': 631}
        with self.assertRaisesRegex(ValidationError, '`Data` object does not exist'):
            Data.objects.create(**data)

    def test_delete_input(self):
        proc1 = Process.objects.create(
            name='Referenced process',
            contributor=self.user,
            type='data:referenced:object:'
        )
        proc2 = Process.objects.create(
            name='Test process',
            contributor=self.user,
            input_schema=[
                {'name': 'data_object', 'type': 'data:referenced:object:'}
            ]
        )
        data1 = Data.objects.create(
            name='Referenced object',
            contributor=self.user,
            process=proc1
        )
        data2 = Data.objects.create(
            name='Test data',
            contributor=self.user,
            process=proc2,
            input={'data_object': data1.pk}
        )

        data1.delete()
        data2.name = 'New name'
        data2.save()


class ValidationUnitTest(TestCase):

    def test_required(self):
        schema = [
            {'name': 'value', 'type': 'basic:integer:', 'required': True},
            {'name': 'description', 'type': 'basic:string:'},  # implicit `required=True`
            {'name': 'comment', 'type': 'basic:string:', 'required': False},
        ]

        instance = {'description': 'test'}
        with self.assertRaisesRegex(ValidationError, '"value" not given.'):
            validate_schema(instance, schema)

        instance = {'value': 42}
        with self.assertRaisesRegex(ValidationError, '"description" not given.'):
            validate_schema(instance, schema)

        instance = {'value': 42, 'description': 'universal answer'}
        validate_schema(instance, schema)

        instance = {'value': 42, 'description': 'universal answer', 'comment': None}
        validate_schema(instance, schema)

        instance = {'value': 42, 'description': 'test', 'comment': 'Lorem ipsum'}
        validate_schema(instance, schema)

        instance = {}
        validate_schema(instance, schema, test_required=False)

    def test_choices(self):
        schema = [
            {
                'name': 'value',
                'type': 'basic:integer:',
                'choices': [{'value': 7, 'label': '7'}, {'value': 13, 'label': '13'}]
            },
        ]

        instance = {'value': 7}
        validate_schema(instance, schema)

        instance = {'value': 8}
        error_msg = "Value of field 'value' must match one of predefined choices. Current value: 8"
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                'name': 'value',
                'type': 'basic:integer:',
                'choices': [{'value': 7, 'label': '7'}, {'value': 13, 'label': '13'}],
                'allow_custom_choice': True
            },
        ]

        instance = {'value': 7}
        validate_schema(instance, schema)

        instance = {'value': 8}
        validate_schema(instance, schema)

    def test_range(self):
        schema = [
            {
                'name': 'value',
                'type': 'basic:integer:',
                'range': [0, 5],
            },
        ]

        instance = {'value': 5}
        validate_schema(instance, schema)
        instance = {'value': 0}
        validate_schema(instance, schema)
        instance = {'value': 3}
        validate_schema(instance, schema)

        instance = {'value': 7}
        error_msg = "Value of field 'value' is out of range. It should be between 0 and 5."
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                'name': 'value',
                'type': 'basic:decimal:',
                'range': [0, 5],
            },
        ]

        instance = {'value': 7}
        error_msg = "Value of field 'value' is out of range. It should be between 0 and 5."
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                'name': 'value',
                'type': 'list:basic:integer:',
                'range': [0, 5],
            },
        ]

        instance = {'value': [0, 3, 5]}
        validate_schema(instance, schema)

        instance = {'value': [0, 3, 7]}
        error_msg = "Value of field 'value' is out of range. It should be between 0 and 5."
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

        schema = [
            {
                'name': 'value',
                'type': 'list:basic:decimal:',
                'range': [0, 5],
            },
        ]

        instance = {'value': [0, 3, 7]}
        error_msg = "Value of field 'value' is out of range. It should be between 0 and 5."
        with self.assertRaisesRegex(ValidationError, error_msg):
            validate_schema(instance, schema)

    def test_missing_in_schema(self):
        schema = [
            {'name': 'result', 'type': 'basic:integer:', 'required': False}
        ]

        instance = {'res': 42}
        with self.assertRaisesRegex(ValidationError, r'\(res\) missing in schema'):
            validate_schema(instance, schema)

    def test_file_prefix(self):
        schema = [
            {'name': 'result', 'type': 'basic:file:'},
        ]
        instance = {'result': {'file': 'result.txt'}}

        data_location = DataLocation.objects.create(subpath='1')

        with patch('resolwe.flow.models.utils.os') as os_mock:
            validate_schema(instance, schema)
            self.assertEqual(os_mock.path.isfile.call_count, 0)

        # missing file
        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.exists = MagicMock(return_value=False)
            with self.assertRaisesRegex(ValidationError, 'Referenced path .* does not exist.'):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.exists.call_count, 1)

        # File is not a normal file
        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.exists = MagicMock(return_value=True)
            os_mock.path.isfile = MagicMock(return_value=False)
            with self.assertRaisesRegex(ValidationError, 'Referenced path .* is not a file.'):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isfile.call_count, 1)

        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.isfile = MagicMock(return_value=True)
            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isfile.call_count, 1)

        instance = {'result': {'file': 'result.txt', 'refs': ['user1.txt', 'user2.txt']}}

        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.isfile = MagicMock(return_value=True)
            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isfile.call_count, 3)

        # missing second `refs` file
        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.isfile = MagicMock(side_effect=[True, True, False])
            os_mock.path.isdir = MagicMock(return_value=False)
            message = 'Path referenced in `refs` .* is neither a file or directory.'
            with self.assertRaisesRegex(ValidationError, message):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isfile.call_count, 3)
            self.assertEqual(os_mock.path.isdir.call_count, 1)

    def test_dir_prefix(self):
        schema = [
            {'name': 'result', 'type': 'basic:dir:'},
        ]
        instance = {'result': {'dir': 'results'}}

        data_location = DataLocation.objects.create(subpath='1')

        # dir validation is not called if `data_location` is not given
        with patch('resolwe.flow.models.utils.os') as os_mock:
            validate_schema(instance, schema)
            self.assertEqual(os_mock.path.isdir.call_count, 0)

        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.exists = MagicMock(return_value=True)
            os_mock.path.isdir = MagicMock(return_value=True)
            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isdir.call_count, 1)

        # missing path
        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.exists = MagicMock(return_value=False)
            with self.assertRaisesRegex(ValidationError, 'Referenced path .* does not exist.'):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.exists.call_count, 1)

        # not a dir
        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.exists = MagicMock(return_value=True)
            os_mock.path.isdir = MagicMock(return_value=False)
            with self.assertRaisesRegex(ValidationError, 'Referenced path .* is not a directory'):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isdir.call_count, 1)

        instance = {'result': {'dir': 'results', 'refs': ['file01.txt', 'file02.txt']}}

        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.isdir = MagicMock(return_value=True)
            os_mock.path.isfile = MagicMock(return_value=True)
            validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isdir.call_count, 1)
            self.assertEqual(os_mock.path.isfile.call_count, 2)

        # missing second `refs` file
        with patch('resolwe.flow.models.utils.os') as os_mock:
            os_mock.path.isdir = MagicMock(side_effect=[True, False])
            os_mock.path.isfile = MagicMock(side_effect=[True, False])
            message = 'Path referenced in `refs` .* is neither a file or directory'
            with self.assertRaisesRegex(ValidationError, message):
                validate_schema(instance, schema, data_location=data_location)
            self.assertEqual(os_mock.path.isdir.call_count, 2)
            self.assertEqual(os_mock.path.isfile.call_count, 2)

    def test_string_field(self):
        schema = [
            {'name': 'string', 'type': 'basic:string:'},
            {'name': 'text', 'type': 'basic:text:'},
        ]

        instance = {'string': 'Test string', 'text': 'Test text'}
        validate_schema(instance, schema)

        instance = {'string': 42, 'text': 42}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_boolean_field(self):
        schema = [
            {'name': 'true', 'type': 'basic:boolean:'},
            {'name': 'false', 'type': 'basic:boolean:'},
        ]

        instance = {'true': True, 'false': False}
        validate_schema(instance, schema)

        instance = {'true': 'true', 'false': 'false'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'true': 1, 'false': 0}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'true': 'foo', 'false': 'bar'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_integer_field(self):
        schema = [
            {'name': 'value', 'type': 'basic:integer:'},
        ]

        instance = {'value': 42}
        validate_schema(instance, schema)

        instance = {'value': 42.0}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'value': 'forty-two'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_decimal_field(self):
        schema = [
            {'name': 'value', 'type': 'basic:decimal:'},
        ]

        instance = {'value': 42}
        validate_schema(instance, schema)

        instance = {'value': 42.0}
        validate_schema(instance, schema)

        instance = {'value': 'forty-two'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_date_field(self):
        schema = [
            {'name': 'date', 'type': 'basic:date:'},
        ]

        instance = {'date': '2000-12-31'}
        validate_schema(instance, schema)

        instance = {'date': '2000/01/01'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'date': '31 04 2000'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'date': '21.06.2000'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'date': '2000-1-1'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'date': '2000 apr 8'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_datetime_field(self):
        schema = [
            {'name': 'date', 'type': 'basic:datetime:'},
        ]

        instance = {'date': '2000-06-21 00:00'}
        validate_schema(instance, schema)

        instance = {'date': '2000 06 21 24:00'}
        with self.assertRaises(ValidationError):
            validate_schema(instance, schema)

        instance = {'date': '2000/06/21 2:03'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'date': '2000-06-21 2:3'}  # XXX: Is this ok?
        validate_schema(instance, schema)

        instance = {'date': '2000-06-21'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'date': '2000-06-21 12pm'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_data_field(self):
        schema = [
            {'name': 'data_list', 'type': 'data:test:upload:'}
        ]
        instance = {
            'data_list': 1
        }

        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.return_value': True,
                'first.return_value': {'process__type': 'data:test:upload:'},
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 1)

        # subtype is OK
        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.return_value': True,
                'first.return_value': {'process__type': 'data:test:upload:subtype:'},
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 1)

        # missing `Data` object
        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.return_value': False,
                'first.return_value': {'process__type': 'data:test:upload:'},
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(ValidationError, '`Data` object does not exist'):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 0)

            validate_schema(instance, schema, skip_missing_data=True)

        # `Data` object of wrong type
        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.return_value': True,
                'first.return_value': {'process__type': 'data:test:wrong:'},
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(ValidationError, 'Data object of type .* is required'):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 1)
            self.assertEqual(value_mock.first.call_count, 1)

        # data `id` shouldn't be string
        instance = {
            'data_list': "1"
        }

        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_file_field(self):
        schema = [
            {'name': 'result', 'type': 'basic:file:', 'validate_regex': r'^.*\.txt$'},
        ]

        instance = {'result': {
            'file': 'result_file.txt',
            'size': 13
        }}
        validate_schema(instance, schema)

        instance = {'result': {
            'file_temp': '12345',
            'file': 'result_file.txt',
        }}
        validate_schema(instance, schema)

        instance = {'result': {
            'file_temp': '12345',
            'is_remote': True,
            'file': 'result_file.txt',
        }}
        validate_schema(instance, schema)

        instance = {'result': {
            'file': 'result_file.txt',
            'refs': ['01.txt', '02.txt'],
        }}
        validate_schema(instance, schema)

        # non-boolean `is_remote`
        instance = {'result': {
            'file_temp': '12345',
            'is_remote': 'ftp',
            'file': 'result_file.txt',
        }}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        # missing `file`
        instance = {'result': {
            'file_temp': '12345',
        }}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        # wrong file extension
        instance = {'result': {
            'file': 'result_file.tar.gz',
        }}
        with self.assertRaisesRegex(ValidationError, 'File name .* does not match regex'):
            validate_schema(instance, schema)

    def test_html_file_field(self):
        schema = [
            {'name': 'html_result', 'type': 'basic:file:html:'},
        ]

        instance = {'html_result': {
            'file': 'index.htmls',
            'refs': ['some.js', 'some.css']
        }}
        validate_schema(instance, schema)

    def test_dir_field(self):
        schema = [
            {'name': 'result', 'type': 'basic:dir:'},
        ]

        instance = {'result': {
            'dir': 'results',
            'size': 32156
        }}
        validate_schema(instance, schema)

        instance = {'result': {
            'dir': 'result',
            'refs': ['01.txt', '02.txt']
        }}
        validate_schema(instance, schema)

        # missing `dir`
        instance = {'result': {
            'refs': ['01.txt', '02.txt'],
        }}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_url_field(self):
        schema = [
            {'name': 'url_view', 'type': 'basic:url:view:', 'required': False},
            {'name': 'url_download', 'type': 'basic:url:download:', 'required': False},
            {'name': 'url_link', 'type': 'basic:url:link:', 'required': False},
        ]

        instance = {
            'url_view': {'url': 'http://www.genialis.com'},
            'url_download': {'url': 'http://www.genialis.com'},
            'url_link': {'url': 'http://www.genialis.com'},
        }
        validate_schema(instance, schema)

        instance = {'url_view': {
            'url': 'http://www.genialis.com',
            'name': 'Genialis',
            'refs': ['http://www.genialis.com/jobs']
        }}
        validate_schema(instance, schema)

        # wrong type
        schema = [
            {'name': 'webpage', 'type': 'basic:url:'},
        ]
        instance = {'webpage': {'url': 'http://www.genialis.com'}}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_json_field(self):
        schema = [
            {'name': 'big_dict', 'type': 'basic:json:'}
        ]

        # json not saved in `Storage`
        instance = {'big_dict': {'foo': 'bar'}}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        with patch('resolwe.flow.models.storage.Storage') as storage_mock:
            filter_mock = MagicMock()
            filter_mock.exists.return_value = True
            storage_mock.objects.filter.return_value = filter_mock

            instance = {'big_dict': 5}
            validate_schema(instance, schema)

            self.assertEqual(filter_mock.exists.call_count, 1)

        # non existing `Storage`
        with patch('resolwe.flow.models.storage.Storage') as storage_mock:
            filter_mock = MagicMock()
            filter_mock.exists.return_value = False
            storage_mock.objects.filter.return_value = filter_mock

            instance = {'big_dict': 5}
            with self.assertRaisesRegex(ValidationError, '`Storage` object does not exist'):
                validate_schema(instance, schema)

            self.assertEqual(filter_mock.exists.call_count, 1)

    def test_list_string_field(self):
        schema = [
            {'name': 'list', 'type': 'list:basic:string:'}
        ]

        instance = {'list': ['foo', 'bar']}
        validate_schema(instance, schema)

        instance = {'list': []}
        validate_schema(instance, schema)

        instance = {'list': ''}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'list': 'foo bar'}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_list_integer_field(self):
        schema = [
            {'name': 'value', 'type': 'list:basic:integer:'},
        ]

        instance = {'value': [42, 43]}
        validate_schema(instance, schema)

        instance = {'value': 42}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'value': [42, 43.0]}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        instance = {'value': [42, "43"]}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_list_data_field(self):
        schema = [
            {'name': 'data_list', 'type': 'list:data:test:upload:'}
        ]
        instance = {
            'data_list': [1, 3, 4]
        }

        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.return_value': True,
                'first.return_value': {'process__type': 'data:test:upload:'},
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 3)
            self.assertEqual(value_mock.first.call_count, 3)

        # subtypes are OK
        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.return_value': True,
                'first.side_effect': [
                    {'process__type': 'data:test:upload:subtype1:'},
                    {'process__type': 'data:test:upload:'},
                    {'process__type': 'data:test:upload:subtype2:'},
                ],
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 3)
            self.assertEqual(value_mock.first.call_count, 3)

        # one object does not exist
        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.side_effect': [True, False, True],
                'first.return_value': {'process__type': 'data:test:upload:'},
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(ValidationError, '`Data` object does not exist'):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 2)
            self.assertEqual(value_mock.first.call_count, 1)

        # one object of wrong type
        with patch('resolwe.flow.models.data.Data') as data_mock:
            value_mock = MagicMock(**{
                'exists.return_value': True,
                'first.side_effect': [
                    {'process__type': 'data:test:upload:'},
                    {'process__type': 'data:test:upload:'},
                    {'process__type': 'data:test:wrong:'},
                ],
            })
            filter_mock = MagicMock(**{'values.return_value': value_mock})
            data_mock.objects.filter.return_value = filter_mock

            with self.assertRaisesRegex(ValidationError, 'Data object of type .* is required'):
                validate_schema(instance, schema)

            self.assertEqual(value_mock.exists.call_count, 3)
            self.assertEqual(value_mock.first.call_count, 3)

    def test_list_file_field(self):
        schema = [
            {'name': 'result', 'type': 'list:basic:file:', 'validate_regex': r'^.*\.txt$'},
        ]
        instance = {'result': [
            {'file': 'result01.txt'},
            {'file': 'result02.txt', 'size': 14, 'refs': ['results.tar.gz']},
        ]}
        validate_schema(instance, schema)

        # wrong extension
        instance = {'result': [
            {'file': 'result01.txt'},
            {'file': 'result02.tar.gz'},
        ]}
        with self.assertRaisesRegex(ValidationError, 'File name .* does not match regex'):
            validate_schema(instance, schema)

    def test_list_dir_field(self):
        schema = [
            {'name': 'result', 'type': 'list:basic:dir:'},
        ]

        instance = {'result': [
            {'dir': 'results01', 'size': 32156, 'refs': ['result01.txt', 'result02.txt']},
            {'dir': 'results02'},
        ]}
        validate_schema(instance, schema)

        # missing `dir`
        instance = {'result': [
            {'size': 32156, 'refs': ['result01.txt', 'result02.txt']},
        ]}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_list_url_field(self):
        schema = [
            {'name': 'webpage', 'type': 'list:basic:url:view:'},
        ]

        instance = {'webpage': [
            {'url': 'http://www.genialis.com', 'refs': ['http://www.genialis.com/jobs']},
            {'url': 'http://www.dictyexpress.org'},
        ]}
        validate_schema(instance, schema)

        instance = {'webpage': {'url': 'http://www.dictyexpress.org'}}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

    def test_groups(self):
        schema = [
            {'name': 'test_group', 'group': [
                {'name': 'result_file', 'type': 'basic:file:', 'validate_regex': r'^.*\.txt$'},
                {'name': 'description', 'type': 'basic:string:', 'required': True}
            ]}
        ]

        instance = {'test_group': {
            'result_file': {'file': 'results.txt'},
            'description': 'This are results',
        }}
        validate_schema(instance, schema)

        # wrong file extension
        instance = {'test_group': {
            'result_file': {'file': 'results.tar.gz'},
            'description': 'This are results',
        }}
        with self.assertRaisesRegex(ValidationError, 'File name .* does not match regex'):
            validate_schema(instance, schema)

        # wrong description type
        instance = {'test_group': {
            'result_file': {'file': 'results.txt'},
            'description': 6,
        }}
        with self.assertRaisesRegex(ValidationError, 'is not valid'):
            validate_schema(instance, schema)

        # missing description
        instance = {'test_group': {
            'result_file': {'file': 'results.txt'},
        }}
        with self.assertRaisesRegex(ValidationError, '"description" not given'):
            validate_schema(instance, schema)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-02-26 04:08
from __future__ import unicode_literals

from django.db import migrations, models


def set_data_relation(apps, schema_editor):
    Data = apps.get_model('flow', 'Data')
    Storage = apps.get_model('flow', 'Storage')

    for data in Data.objects.all():
        storage = Storage.objects.filter(data_migration_temporary=data).first()
        if storage:
            storage.data.add(data)


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0028_add_data_location'),
    ]

    operations = [
        migrations.RenameField(
            model_name='storage',
            old_name='data',
            new_name='data_migration_temporary',
        ),
        migrations.AddField(
            model_name='storage',
            name='data',
            field=models.ManyToManyField(related_name='storages', to='flow.Data'),
        ),
        migrations.RunPython(set_data_relation),
        migrations.RemoveField(
            model_name='storage',
            name='data_migration_temporary',
        ),
    ]



"""Resolwe model database functions."""
from django.db.models.aggregates import Func

from .utils import json_path_components


class JsonGetPath(Func):  # pylint: disable=abstract-method
    """PostgreSQL JSON path (#>) operator."""

    function = '#>'
    template = "%(expressions)s%(function)s%%s"
    arity = 1

    def __init__(self, expression, path):
        """Initialize function.

        :param expression: Expression that returns a JSON field
        :param path: Path to get inside the JSON object, which can be
            either a list of path components or a dot-separated
            string
        """
        self.path = json_path_components(path)

        super().__init__(expression)

    def as_sql(self, compiler, connection):  # pylint: disable=arguments-differ
        """Compile SQL for this function."""
        sql, params = super().as_sql(compiler, connection)
        params.append(self.path)
        return sql, params


class JsonbArrayElements(Func):  # pylint: disable=abstract-method
    """PostgreSQL jsonb_array_elements function."""

    function = 'jsonb_array_elements'
    template = '%(function)s(%(expressions)s)'
    arity = 1



""".. Ignore pydocstyle D400.

===================
Workload Connectors
===================

The workload management system connectors are used as glue between the
Resolwe Manager and various concrete workload management systems that
might be used by it. Since the only functional requirement is job
submission, they can be simple and nearly contextless.

.. automodule:: resolwe.flow.managers.workload_connectors.base
    :members:
.. automodule:: resolwe.flow.managers.workload_connectors.local
    :members:
.. automodule:: resolwe.flow.managers.workload_connectors.celery
    :members:
.. automodule:: resolwe.flow.managers.workload_connectors.slurm
    :members:

"""



"""Collection viewset."""
from distutils.util import strtobool  # pylint: disable=import-error,no-name-in-module

from elasticsearch_dsl.query import Q

from django.db.models.query import Prefetch

from rest_framework import exceptions, mixins, status, viewsets
from rest_framework.decorators import detail_route, list_route
from rest_framework.response import Response

from resolwe.elastic.composer import composer
from resolwe.elastic.viewsets import ElasticSearchCombinedViewSet
from resolwe.flow.models import Collection, Data
from resolwe.flow.serializers import CollectionSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.permissions.utils import remove_permission, update_permission

from ..elastic_indexes import CollectionDocument
from .mixins import ParametersMixin, ResolweCheckSlugMixin, ResolweCreateModelMixin, ResolweUpdateModelMixin


class CollectionViewSet(ElasticSearchCombinedViewSet,
                        ResolweCreateModelMixin,
                        mixins.RetrieveModelMixin,
                        ResolweUpdateModelMixin,
                        mixins.DestroyModelMixin,
                        mixins.ListModelMixin,
                        ResolwePermissionsMixin,
                        ResolweCheckSlugMixin,
                        ParametersMixin,
                        viewsets.GenericViewSet):
    """API view for :class:`Collection` objects."""

    queryset = Collection.objects.all().prefetch_related(
        'descriptor_schema',
        'contributor',
        Prefetch('data', queryset=Data.objects.all().order_by('id'))
    )
    serializer_class = CollectionSerializer
    permission_classes = (get_permissions_class(),)
    document_class = CollectionDocument

    filtering_fields = (
        'id', 'slug', 'name', 'created', 'modified', 'contributor', 'owners', 'text', 'tags',
    )
    filtering_map = {
        'name': 'name.raw',
        'contributor': 'contributor_id',
        'owners': 'owner_ids',
    }
    ordering_fields = ('id', 'created', 'modified', 'name', 'contributor')
    ordering_map = {
        'name': 'name.raw',
        'contributor': 'contributor_sort',
    }
    ordering = 'id'

    def get_queryset(self):
        """Return queryset."""
        if self.request and self.request.query_params.get('hydrate_data', False):
            return self.queryset.prefetch_related('entity_set')

        return self.queryset

    def custom_filter_tags(self, value, search):
        """Support tags query."""
        if not isinstance(value, list):
            value = value.split(',')

        filters = [Q('match', **{'tags': item}) for item in value]
        search = search.query('bool', must=filters)

        return search

    def get_always_allowed_arguments(self):
        """Return query arguments which are always allowed."""
        return super().get_always_allowed_arguments() + [
            'hydrate_data',
        ]

    def custom_filter_text(self, value, search):
        """Support general query using the 'text' attribute."""
        if isinstance(value, list):
            value = ' '.join(value)

        should = [
            Q('match', slug={'query': value, 'operator': 'and', 'boost': 10.0}),
            Q('match', **{'slug.ngrams': {'query': value, 'operator': 'and', 'boost': 5.0}}),
            Q('match', name={'query': value, 'operator': 'and', 'boost': 10.0}),
            Q('match', **{'name.ngrams': {'query': value, 'operator': 'and', 'boost': 5.0}}),
            Q('match', contributor_name={'query': value, 'operator': 'and', 'boost': 5.0}),
            Q('match', **{'contributor_name.ngrams': {'query': value, 'operator': 'and', 'boost': 2.0}}),
            Q('match', owner_names={'query': value, 'operator': 'and', 'boost': 5.0}),
            Q('match', **{'owner_names.ngrams': {'query': value, 'operator': 'and', 'boost': 2.0}}),
            Q('match', descriptor_data={'query': value, 'operator': 'and'}),
        ]

        # Add registered text extensions.
        for extension in composer.get_extensions(self):
            if hasattr(extension, 'text_filter'):
                should += extension.text_filter(value)

        search = search.query('bool', should=should)

        return search

    def set_content_permissions(self, user, obj, payload):
        """Apply permissions to data objects and entities in ``Collection``."""
        for entity in obj.entity_set.all():
            if user.has_perm('share_entity', entity):
                update_permission(entity, payload)

        # Data doesn't have "ADD" permission, so it has to be removed
        payload = remove_permission(payload, 'add')

        for data in obj.data.all():
            if user.has_perm('share_data', data):
                update_permission(data, payload)

    def create(self, request, *args, **kwargs):
        """Only authenticated usesr can create new collections."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        return super().create(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        """Destroy a model instance.

        If ``delete_content`` flag is set in query parameters, also all
        Data objects and Entities, on which user has ``EDIT``
        permission, contained in collection will be deleted.
        """
        obj = self.get_object()
        user = request.user

        if strtobool(request.query_params.get('delete_content', 'false')):
            for entity in obj.entity_set.all():
                if user.has_perm('edit_entity', entity):
                    entity.delete()

            for data in obj.data.all():
                if user.has_perm('edit_data', data):
                    data.delete()

        return super().destroy(request, *args, **kwargs)  # pylint: disable=no-member

    @detail_route(methods=['post'])
    def add_data(self, request, pk=None):
        """Add data to collection."""
        collection = self.get_object()

        if 'ids' not in request.data:
            return Response({"error": "`ids`parameter is required"}, status=status.HTTP_400_BAD_REQUEST)

        missing = []
        for data_id in request.data['ids']:
            if not Data.objects.filter(pk=data_id).exists():
                missing.append(data_id)

        if missing:
            return Response(
                {"error": "Data objects with following ids are missing: {}".format(', '.join(missing))},
                status=status.HTTP_400_BAD_REQUEST)

        for data_id in request.data['ids']:
            collection.data.add(data_id)

        return Response()

    @detail_route(methods=['post'])
    def remove_data(self, request, pk=None):
        """Remove data from collection."""
        collection = self.get_object()

        if 'ids' not in request.data:
            return Response({"error": "`ids`parameter is required"}, status=status.HTTP_400_BAD_REQUEST)

        for data_id in request.data['ids']:
            collection.data.remove(data_id)

        return Response()

    @list_route(methods=['post'])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Collection`` models."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        ids = self.get_ids(request.data)
        queryset = get_objects_for_user(request.user, 'view_collection', Collection.objects.filter(id__in=ids))
        actual_ids = queryset.values_list('id', flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Collections with the following ids not found: {}".format(', '.join(map(str, missing_ids)))
            )

        duplicated = queryset.duplicate(contributor=request.user)

        serializer = self.get_serializer(duplicated, many=True)
        return Response(serializer.data)
